/*
Copyright 2011, George Smart, M1GEO (http://www.george-smart.co.uk). All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY GEORGE SMART "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GEORGE SMART OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of George Smart.
*/

// Compile with
// gcc -Wall -lk8055 -o "main" "main.c"

/* This code has an option 'i2c_delay_ms' to allow the I2C device to 
 * react to commands, etc.  However, as the IO on the Velleman K8055 is
 * SLOOOOOW, we don't really need it.  It is useful for slowing the writing
 * down, however, or if you wish to port this to anything else?!
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <math.h>

#include <k8055.h>

#define	FALSE			0
#define	TRUE			!FALSE

#define i2c_delay_ms	0		// microseconds
#define	i2c_sda_value	1
#define i2c_scl_value	2
#define i2c_busy_led	128

#define	i2c_slave_add	0x70
#define	i2c_slave_sta	0x00
#define	i2c_slave_ctl	0xF7

int		iCardAddress = 0;
long 	iDigitalOutputValues = 0;

char	character_map[] = {	0b11111100,	//0
							0b01100000,	//1
							0b11011010,	//2
							0b11110010,	//3
							0b01100110,	//4
							0b10110110,	//5
							0b10111110,	//6
							0b11100000,	//7
							0b11111110,	//8
							0b11100110,	//9
							0b11101100,	//A
							0b00111110,	//B
							0b10011100,	//C
							0b01111010,	//D
							0b10011110,	//E
							0b10001110,	//F
							0b00000001,	//D.P.
							0b00000000	//BLANKD
										};

/* As we can't read the output bytes back, nor can we change individual bits, we must
 * keep a register of what bits are what.  This MUST be kept up to date. ALWAYS!
 * 
 * Such is the purpose of the 'iDigitalOutputValues' integer.
 * 
 * Also note, on the Velleman K8055, setting a 1 means the output is shorted low.
 * Hardware inversion of the signal: setting 1 sets the base of the ULA2803, and
 * so the bus is shorted to 0.
 */

void BUSY(int bit) {
	if (bit>0) {	// dont invert as we want the LED
		iDigitalOutputValues = iDigitalOutputValues | i2c_busy_led; // set SDA bit in memory
	} else {
		iDigitalOutputValues = iDigitalOutputValues & ~i2c_busy_led;        // clear SDA bit in memory
	}
	WriteAllDigital(iDigitalOutputValues);	// send changes to hardware
}

void SDA(int bit) {
	if (bit<1) {
		iDigitalOutputValues = iDigitalOutputValues | i2c_sda_value; // set SDA bit in memory
	} else {
		iDigitalOutputValues = iDigitalOutputValues & ~i2c_sda_value;        // clear SDA bit in memory
	}
	WriteAllDigital(iDigitalOutputValues);	// send changes to hardware
}

void SCL(int bit) {
	if (bit<1) {
		iDigitalOutputValues = iDigitalOutputValues | i2c_scl_value; // set SDA bit in memory
	} else {
		iDigitalOutputValues = iDigitalOutputValues & ~i2c_scl_value;        // clear SDA bit in memory
	}
	WriteAllDigital(iDigitalOutputValues);	// send changes to hardware
}

void i2c_start() {
	BUSY(1);
	SDA(1);
	SCL(1);
	usleep(i2c_delay_ms/2);
	SDA(0);
	usleep(i2c_delay_ms/2);
	SCL(0);
}

void i2c_write(unsigned char byte) {
	int i = 0;
	for (i = 0; i < 8; i++) {
		SCL(0);
		if (byte & 0x80) { SDA(1); } else { SDA(0); }
		usleep(i2c_delay_ms/2);
		SCL(1);
		usleep(i2c_delay_ms/2);
		byte <<= 1;
	}
	SCL(0);
	SDA(1);
	usleep(i2c_delay_ms/2);
	SCL(1);	// provide one clock cycle for the ack bit (which we ignore)
	usleep(i2c_delay_ms/2);
}

void i2c_stop() {
	SCL(0);
	SDA(0);
	usleep(i2c_delay_ms/2);
	SCL(1);
	usleep(i2c_delay_ms/2);
	SDA(1);
	BUSY(0);
}

void set_digits(unsigned char d1, unsigned char d2, unsigned char d3, unsigned char d4, unsigned char point) {
	i2c_start();		// start i2c
	i2c_write(i2c_slave_add);	// address
	i2c_write(i2c_slave_sta);	// instruction
	i2c_write(i2c_slave_ctl);	// control
	i2c_write(character_map[d1]);		// digit 1
	i2c_write(character_map[d2]+point);	// digit 2
	i2c_write(character_map[d3]);		// digit 3
	i2c_write(character_map[d4]);		// digit 4
	i2c_stop();			// stop i2c
}

unsigned char hex2bcd (unsigned char x, int a) {
    unsigned char y = 0, z = 0;
    y = (x / 10);
    z = z | (x % 10);
    if (a){
		return (y);	// return tens
	} else {
		return (z);	// return units
	}
}

int main(int argc, char **argv)
{
	
	printf("Velleman K8055 I2C System\n");
	printf("George Smart, M1GEO\n\n");
	
	
	// Open Card...
	printf("Connecting to K8055:%d: ", iCardAddress);
	iCardAddress = OpenDevice((long) iCardAddress);
	if (iCardAddress >= 0) {
		printf("Connected.\n\n");
	} else {
		printf("Failed!\n\n");
		exit(EXIT_FAILURE);
	}

	// float the bus - idle condition.
	SetAllAnalog();	// set both analog to maximum (used here as 5V pullup).

	struct tm  *ts;
	time_t now;
	int h[] = {1,2};
	int m[] = {3,4};
	
	int old_min = -1;
	int new_min = -2;

	while(TRUE) {
		now = time(NULL);
		ts = localtime(&now);
		h[1]=hex2bcd(ts->tm_hour, 1);
		h[0]=hex2bcd(ts->tm_hour, 0);
		m[1]=hex2bcd(ts->tm_min, 1);
		m[0]=hex2bcd(ts->tm_min, 0);
		new_min = ts->tm_min;
		if ((old_min == new_min) == FALSE) {
			printf("\r%02d:%02d:%02d - Updating ", ts->tm_hour, ts->tm_min, ts->tm_sec); fflush(stdout);
			set_digits(h[1], h[0], m[1], m[0], 1);
		} else {
			printf("\r%02d:%02d:%02d - Waiting  ", ts->tm_hour, ts->tm_min, ts->tm_sec); fflush(stdout);
			usleep (1000000);
		}
		old_min = ts->tm_min;
	}

	printf("Closing K8055 at Address %d: ", iCardAddress);
	CloseDevice();
	printf("Closed.\n");
	
	return 0;
}



